from enum import Enum, auto
from typing import Optional
from arkitekt.config import TransportProtocol
from herre.access.object import GraphQLObject



class AgentSettings(GraphQLObject):
    type: TransportProtocol
    kwargs: dict

class PostmanSettings(GraphQLObject):
    type: TransportProtocol
    kwargs: dict

class Transcript(GraphQLObject):
    postman: Optional[PostmanSettings]
    agent: Optional[AgentSettings]



import argparse
from enum import Enum
from genericpath import isfile
from runpy import run_path
from arkitekt.cli.dev.autoreload import watch_directory_and_restart
from fakts import Fakts
from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.beacon import BeaconGrant
from fakts.grants.endpoint import EndpointGrant
from fakts.grants.cli.clibeacon import CLIBeaconGrant
from fakts.grants.yaml import YamlGrant
from herre import Herre
from rich.console import Console
from rich.prompt import Prompt, Confirm
import os
import asyncio

directory = os.getcwd()


class ArkitektOptions(str, Enum):
    INIT = "init"
    DEV = "dev"
    LOGOUT = "logout"
    LOGIN = "login"
    RUN = "run"


agent_script = f'''
from arkitekt.agents.script import ScriptAgent

agent = ScriptAgent()

@agent.register()
def test_show(name: str)-> str:
    """Demo Node

    This demo node will just pass through
    a name 

    Args:
        name (str): The name

    Returns:
        str: The pass throughed Name
    """
    return name

agent.provide()

'''

mikro_script = """
from mikro import gql

x = gql('''
myrepresentations {
    id
}
''')

"""


template_map = {"agent": agent_script, "client": mikro_script}


def main(
    script=ArkitektOptions.INIT,
    name=None,
    path=".",
    refresh=False,
    scan_network=None,
    beacon=None,
    template=None,
):

    console = Console()

    if path == ".":
        app_directory = os.getcwd()
        name = name or os.path.basename(app_directory)
    else:
        app_directory = os.path.join(os.getcwd(), path)
        name = name or path

    fakts_path = os.path.join(app_directory, "fakts.yaml")
    run_script_path = os.path.join(app_directory, "run.py")

    if script == ArkitektOptions.DEV:

        if not os.path.isfile(run_script_path):
            console.print(f"Could't find a run.py in {app_directory} {run_script_path}")
            return
        if not os.path.isfile(fakts_path):
            console.print(f"{app_directory} does not have a valid fakts.yaml")
            return

        asyncio.run(watch_directory_and_restart(app_directory, run_script_path))

    if script == ArkitektOptions.RUN:

        if not os.path.isfile(run_script_path):
            console.print(f"Could't find a run.py in {app_directory} {run_script_path}")
            return
        if not os.path.isfile(fakts_path):
            console.print(f"{app_directory} does not have a valid fakts.yaml")
            return

        run_path(run_script_path)

    if script == ArkitektOptions.LOGIN:

        if not os.path.isfile(fakts_path):
            console.print(
                f"Directory does not containt a valid fakts.yaml. Please initialize through 'arkitekt init {path}' first"
            )
            return

        fakts = Fakts(grants=[], fakts_path=fakts_path)
        if not fakts.loaded:
            console.print(
                f"Configuration in fakts.yaml is not sufficient please reinitilize through 'arkitekt init {path}'"
            )

        herre = Herre(fakts=fakts)
        console.print("Loggin in")
        herre.login()

    if script == ArkitektOptions.LOGOUT:

        if not os.path.isfile(fakts_path):
            console.print(
                f"Directory does not contain a valid fakts.yaml. If you want to login again please reinitiliaze."
            )
            return

        fakts = Fakts(grants=[], fakts_path=fakts_path)
        if not fakts.loaded:
            console.print(
                f"Configuration in fakts.yaml is not sufficient please reinitilize through 'arkitekt init {path}'"
            )

        herre = Herre(fakts=fakts)
        console.print("Logging out!")
        herre.logout()

    if script == ArkitektOptions.INIT:

        console.print("Initializing Arkitekt Started. Lets do this!")

        if os.path.exists(app_directory):
            console.print(f"App Directory already existed. No need for creation!")
        else:
            os.mkdir(app_directory)
            console.print(f"Created a new App Directory at {app_directory}")

        initialize_fakts = True
        if os.path.isfile(fakts_path):
            console.print(
                f"There seems to have been a previous App initialized at {app_directory}"
            )
            if refresh is None:
                initialize_fakts = Confirm.ask(
                    "Do you want to refresh the Apps configuration? :smiley: "
                )
            else:
                initialize_fakts = refresh

        if initialize_fakts:

            hard_fakts = {"herre": {"name": name}}

            fakt_grants = []

            console.print("---------------------------------------------")
            console.print(f"Initializing a new Configuration for {name}")

            if scan_network is None:
                scan_network = Confirm.ask(
                    "Do you want to automatically scan the network for local instances?"
                )
                if scan_network:
                    fakt_grants.append(CLIBeaconGrant())

            if scan_network is False and beacon is None:
                beacon = Prompt.ask(
                    "Give us your local beacon",
                    default="http://localhost:3000/setupapp",
                )
                fakt_grants.append(
                    EndpointGrant(FaktsEndpoint(url=beacon, name="local_beacon"))
                )

            fakts = Fakts(
                grants=fakt_grants,
                fakts_path=fakts_path,
                force_reload=initialize_fakts,
                hard_fakts=hard_fakts,
            )

            if not fakts.loaded:
                fakts.load()

            console.print("--------------------------------------------")

        save_run = True
        if os.path.isfile(run_script_path):
            console.print("Initializing Templates for the App")
            save_run = Confirm.ask(
                "run.py already exists? Do you want to overwrite the run.py file? :smiley: "
            )

        if save_run:
            if not template:
                template = Prompt.ask(
                    "Which Template do you want to use?",
                    choices=list(template_map.keys()),
                    default="agent",
                )

            with open(run_script_path, "w") as f:
                f.write(template_map[template])
                console.print("Create new Entrypoint for this App")


def entrypoint():
    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("script", type=ArkitektOptions, help="The Script Type")
    parser.add_argument("path", type=str, help="The Path", nargs="?", default=".")
    parser.add_argument("--name", type=str, help="The Name of this script")
    parser.add_argument("--refresh", type=bool, help="Do you want to refresh")
    parser.add_argument("--scan-network", type=bool, help="Do you want to refresh")
    parser.add_argument("--beacon", type=str, help="The adress of the beacon")
    parser.add_argument("--template", type=str, help="The run script template")
    args = parser.parse_args()

    main(
        script=args.script,
        path=args.path,
        name=args.name,
        refresh=args.refresh,
        scan_network=args.scan_network,
        template=args.template,
    )


if __name__ == "__main__":
    entrypoint()



import asyncio
import sys
import time
import logging
from rich.console import Console
from watchdog.observers import Observer
from watchdog.events import (
    FileModifiedEvent,
    FileSystemEventHandler,
    LoggingEventHandler,
)
import os
import subprocess
import janus
import threading
import os
import signal


class QueueHandler(FileSystemEventHandler):
    def __init__(self, *args, sync_q=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = sync_q

    def on_any_event(self, event):
        self.queue.put(event)
        self.queue.join()


def watcher(path, queue, event):
    event_handler = QueueHandler(sync_q=queue)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()

    while True:
        time.sleep(1)


async def buffered_queue(queue, timeout=4):
    buffer = [None]

    async def iterator():
        while True:
            nana = await queue.get()
            queue.task_done()
            if isinstance(nana, FileModifiedEvent):
                if nana.src_path.lower().endswith((".py")):
                    buffer.append(nana)

    loop = asyncio.get_event_loop()
    loop.create_task(iterator())
    while True:
        await asyncio.sleep(1)
        yield buffer[len(buffer) - 1]
        buffer = [None]


async def _read_stream(stream, cb):
    while True:
        line = await stream.readline()
        if line:
            cb(line.decode())
        else:
            break


async def run_subproc(cmd, outcb=print, errorcb=print, env=None):

    try:
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env
        )

        await asyncio.wait(
            [_read_stream(proc.stdout, outcb), _read_stream(proc.stderr, errorcb)]
        )
        await proc.wait()
        print("Program Failed. Please Correct Error!")
        return

    except asyncio.CancelledError as e:
        proc.kill()
        proc.terminate()
        await proc.wait()
        raise e


class Host:
    def __init__(
        self,
        path=None,
        run_py="run.py",
        outcb=sys.stdout.write,
        errorcb=sys.stderr.write,
    ) -> None:
        self.script_path = os.path.join(path, run_py)
        self.interpreter_path = sys.executable
        self.process_task = None
        self.console = Console()
        self.outcb = self.console.print
        self.errorcb = self.console.print
        self.env = os.environ.copy()
        self.env["PYTHONUNBUFFERED"] = "1"
        self.env["ARKITEKT_AGENT_DEBUG"] = "True"

    async def restart(self):
        loop = asyncio.get_event_loop()
        if self.process_task:
            print("Cancelling our Initial Task")
            self.process_task.cancel()
            try:
                await self.process_task
            except asyncio.CancelledError as e:
                print("Process Completely Cancelled")

        print("Starting new Task")
        self.process_task = loop.create_task(
            run_subproc(
                [self.interpreter_path, self.script_path],
                outcb=self.outcb,
                errorcb=self.errorcb,
                env=self.env,
            )
        )

    async def run(self, queue):
        await self.restart()

        async for event in buffered_queue(queue, timeout=1):
            if event:
                await self.restart()


async def watch_directory_and_restart(path="watch", run_py="run.py"):

    cancel_event = threading.Event()
    jqueue = janus.Queue()
    loop = asyncio.get_running_loop()
    fut = loop.run_in_executor(None, watcher, path, jqueue.sync_q, cancel_event)

    host = Host(path=path, run_py=run_py)
    await host.run(jqueue.async_q)
    await fut

    jqueue.close()
    await jqueue.wait_closed()



from arkitekt.packers.structure import Structure, StructureDefaults
from pydantic import BaseModel

from arkitekt.schema.widgets import Widget


class StructureModel(Structure):
    """Model

    Model is the abstract baseclass of all Serverside Models and provides a Django ORM
    like interface for retrieving data from the Server.

    Implements:
        id (str): Every Model has an id (UUID) that identifies the Server Instance

    Args:
        BaseModel ([type]): [description]
        Structure ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ModelMeta.

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """

    @classmethod
    def get_identifier(cls):
        return cls.get_meta().identifier

    @classmethod
    def get_defaults(cls) -> StructureDefaults:
        meta = cls.get_meta()
        return StructureDefaults(widget=getattr(meta, "default_widget", None))



from arkitekt.transport.base import Transport, TransportConfig


class PostmanTransportConfig(TransportConfig):
    pass


class PostmanTransport(Transport):
    configClass = PostmanTransportConfig
    config: PostmanTransportConfig



from abc import ABC, abstractclassmethod, abstractmethod
from typing import AbstractSet, Union

from pydantic.main import BaseModel

from arkitekt.schema.widgets import Widget


class StructureDefaults(BaseModel):
    widget: Widget


class Structure(ABC):
    """An Abstract Mixin to enforce packaging as a Structure before registering

    Args:
        ABC ([type]): [description]

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """

    @abstractclassmethod
    def get_identifier(cls):
        raise NotImplementedError(
            "Every Structure needs to specify how to get its identifier"
        )

    @abstractclassmethod
    def get_defaults(cls) -> StructureDefaults:
        return StructureDefaults()

    @abstractclassmethod
    async def expand(cls, shrinked_value):
        raise NotImplementedError("Every Structure needs to implement a expand Method")

    @abstractmethod
    async def shrink(self):
        raise NotImplementedError("Every Structure needs to implement a shrink Method")



from abc import abstractmethod
from arkitekt.agents.base import Agent
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.messages.postman.unprovide.bounced_unprovide import (
    BouncedUnprovideMessage,
)
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.messages.postman.provide.provide_transition import (
    ProvideState,
    ProvideTransitionMessage,
)
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage


import logging

logger = logging.getLogger(__name__)


class StandardAgent(Agent):
    @abstractmethod
    async def on_bounced_provide(message: BouncedProvideMessage):
        raise NotImplementedError()

    @abstractmethod
    async def on_bounced_unprovide(message: BouncedUnprovideMessage):
        raise NotImplementedError()

    @abstractmethod
    async def on_bounced_assign(message: BouncedForwardedAssignMessage):
        raise NotImplementedError()

    @abstractmethod
    async def on_bounced_unassign(message: BouncedForwardedUnassignMessage):
        raise NotImplementedError()

    async def handle_bounced_provide(self, message: BouncedProvideMessage):
        try:
            print("Received ", message)
            await self.on_bounced_provide(message)
            await self.transport.forward(
                ProvideLogMessage.from_log(
                    message.meta.reference,
                    "Forwarded to Agent",
                    extensions=message.meta.extensions,
                )
            )
        except Exception as e:
            logger.error(e)
            await self.transport.forward(
                ProvideTransitionMessage.from_critical(
                    message.meta.reference, e, extensions=message.meta.extensions
                )
            )

    async def handle_bounced_unprovide(self, message: BouncedUnprovideMessage):
        try:
            await self.on_bounced_unprovide(message)
            await self.transport.forward(
                ProvideLogMessage.from_log(
                    message.data.provision,
                    "Actor Delation Happening",
                    extensions=message.meta.extensions,
                )
            )
        except Exception as e:
            logger.error(e)
            await self.transport.forward(
                ProvideTransitionMessage.from_critical(
                    message.data.provision, e, extensions=message.meta.extensions
                )
            )

    async def handle_bounced_assign(self, message: BouncedForwardedAssignMessage):
        try:
            await self.on_bounced_assign(message)
            await self.transport.forward(
                AssignLogMessage.from_log(
                    message.meta.reference,
                    "Forwared Assignment",
                    extensions=message.meta.extensions,
                )
            )

        except Exception as e:
            await self.transport.forward(
                AssignCriticalMessage.from_critical(
                    message.meta.reference, e, extensions=message.meta.extensions
                )
            )
            raise e

    async def handle_bounced_unassign(self, message: BouncedForwardedUnassignMessage):
        try:
            await self.on_bounced_unassign(message)
            await self.transport.forward(
                AssignLogMessage.from_log(
                    message.data.assignation,
                    "Received Cancellation",
                    extensions=message.meta.extensions,
                )
            )

        except Exception as e:
            logger.error(e)
            await self.transport.forward(
                AssignCriticalMessage.from_critical(
                    message.data.assignation, e, extensions=message.meta.extensions
                )
            )
            raise e



from qtpy import QtWidgets, QtGui
from arkitekt.qt.images.dir import get_image_path
from arkitekt.qt.widgets.settings_popup import SettingsPopup
from fakts.qt import QtFakts
from herre.qt import QtHerre
from arkitekt.qt.agent import QtAgent
import traceback


class MagicBar(QtWidgets.QWidget):
    settingsPopupClass = SettingsPopup

    def __init__(self, fakts: QtFakts, herre: QtHerre, agent: QtAgent, *args, parent=None, darkMode=False, **kwargs) -> None:
        super().__init__(*args, parent=parent,**kwargs)

        self.darkMode = darkMode
        self.fakts = fakts
        self.herre = herre
        self.agent = agent

        self.fakts.loaded_signal.connect(self.on_facts_loaded)
        self.herre.login_signal.connect(self.on_herre_login)
        self.agent.provide_signal.connect(self.on_agent_provide)

        # Tasks 
        self.load_fakts_task = None
        self.herre_login_task = None
        self.agent_provide_task = None

        #Settings
        self.gear_button_popup = self.settingsPopupClass(self)


        self.magicb = QtWidgets.QPushButton("Connect")
        self.magicb.setMinimumHeight(30)
        self.magicb.setMaximumHeight(30)

        self.layout = QtWidgets.QHBoxLayout()
        self.gearb_pix = QtGui.QPixmap(get_image_path("gear.png", darkMode=darkMode))
        self.gearb = QtWidgets.QPushButton()
        self.gearb.setIcon(QtGui.QIcon(self.gearb_pix))
        self.gearb.setMinimumWidth(30)
        self.gearb.setMaximumWidth(30)
        self.gearb.setMinimumHeight(30)
        self.gearb.setMaximumHeight(30)


        self.layout.addWidget(self.magicb)
        self.layout.addWidget(self.gearb)
        self.setLayout(self.layout)


        self.magicb.clicked.connect(self.magic_button_clicked)
        self.gearb.clicked.connect(self.gear_button_clicked)



        if not self.fakts.loaded:
            self.set_unkonfigured()

        else:
            self.set_unconnected()


    def on_agent_except(self, exc_obj):
        self.set_halted()
        errorbox = QtWidgets.QMessageBox()
        errorbox.setText(''.join(traceback.format_exception(None, exc_obj, exc_obj.__traceback__)))
        errorbox.exec_()

    def on_fakts_except(self, exc_obj):
        self.set_unkonfigured()
        errorbox = QtWidgets.QMessageBox()
        errorbox.setText(''.join(traceback.format_exception(None, exc_obj, exc_obj.__traceback__)))
        errorbox.exec_()

    def on_herre_except(self, exc_obj):
        self.set_unconnected()
        errorbox = QtWidgets.QMessageBox()
        errorbox.setText(''.join(traceback.format_exception(None, exc_obj, exc_obj.__traceback__)))
        errorbox.exec_()


    def on_facts_loaded(self, konfik_state):
        if konfik_state:  self.set_unconnected()
        else: self.set_unkonfigured()
    
    def on_herre_login(self, login_state):
        if login_state: self.set_unprovided()
        else: self.set_unconnected()

    def on_agent_provide(self, provide_state):
        if provide_state: self.set_providing()
        else: self.set_halted()


    def gear_button_clicked(self):
        self.gear_button_popup.show()

    def magic_button_clicked(self):
        if not self.fakts.loaded:
            # App is not Konfigured Yet lets do this
            if self.load_fakts_task: 
                self.load_fakts_task.cancel()
                self.load_fakts_task = None

            self.load_fakts_task = self.fakts.load(as_task=True)
            self.load_fakts_task.except_signal.connect(self.on_fakts_except)
            return 

        if not self.herre.logged_in:
            if self.herre_login_task: 
                self.herre_login_task.cancel()
                self.herre_login_task = None

            self.herre_login_task = self.herre.login(as_task=True) 
            self.herre_login_task.except_signal.connect(self.on_herre_except) 
            return 

        if not self.agent_provide_task:
            self.agent_provide_task = self.agent.provide(as_task=True)
            self.agent_provide_task.except_signal.connect(self.on_agent_except)
            return
        
        else:
            self.agent_provide_task.cancel()
            self.agent_provide_task = None
            return 

        
    def update_movie(self):
        self.magicb.setIcon(QtGui.QIcon(self.magicb_movie.currentPixmap()))

    def set_unkonfigured(self):
        self.magicb_movie = QtGui.QMovie(get_image_path("pink pulse.gif", darkMode=self.darkMode))
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.start()

        self.magicb.setText("Konfigure App")

    def set_unconnected(self):
        self.magicb_movie = QtGui.QMovie(get_image_path("green pulse.gif", darkMode=self.darkMode))
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.start()

        self.magicb.setText("Connect")

    def set_halted(self):
        self.magicb_movie = QtGui.QMovie(get_image_path("orange pulse.gif", darkMode=self.darkMode))
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.start()

        self.magicb.setText("Halted")

    def set_unprovided(self):
        self.magicb_movie = QtGui.QMovie(get_image_path("green pulse.gif", darkMode=self.darkMode))
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.start()

        self.magicb.setText("Provide")

    def set_providing(self):
        self.magicb_movie = QtGui.QMovie(get_image_path("green pulse.gif", darkMode=self.darkMode))
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.start()

        self.magicb.setText("Providing...")





from typing import List, Optional

from pydantic.main import BaseModel


class ReserveParams(BaseModel):
    providers: Optional[List[str]]
    templates: Optional[List[str]]
    auto_unprovide: Optional[bool] = True
    auto_provide: Optional[bool] = True


from pydantic.main import BaseModel
from ....messages.types import  RESERVE_ACTIVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = RESERVE_ACTIVE
    extensions: Optional[MetaExtensionsModel]

class DataModel(MessageDataModel):
    topic: Optional[str] #TODO: Maybe not optional

class ReserveActiveMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from enum import Enum
from os import strerror
from arkitekt.packers.transpilers.base import Transpiler
from arkitekt.actors.base import Actor
from arkitekt.packers.structure import Structure
from arkitekt.schema.node import Node
from arkitekt.schema.enums import NodeType
from arkitekt.schema.ports import (
    ArgPort,
    BoolArgPort,
    BoolKwargPort,
    BoolReturnPort,
    DictArgPort,
    DictKwargPort,
    DictReturnPort,
    EnumArgPort,
    EnumKwargPort,
    IntArgPort,
    IntKwargPort,
    IntReturnPort,
    KwargPort,
    ListArgPort,
    ListKwargPort,
    ListReturnPort,
    ReturnPort,
    StringArgPort,
    StringKwargPort,
    StringReturnPort,
    StructureArgPort,
    StructureKwargPort,
    StructureReturnPort,
)
from arkitekt.actors.functional import (
    FunctionalFuncActor,
    FunctionalGenActor,
    FunctionalThreadedFuncActor,
    FunctionalThreadedGenActor,
)
from arkitekt.packers.transpilers.registry import get_transpiler_registry

import inspect
from inspect import signature, Parameter
from docstring_parser import parse
import logging
import inflection

logger = logging.getLogger(__name__)
from typing import Callable, Dict, List
import inspect


class ConversionError(Exception):
    pass


class NonConvertableType(ConversionError):
    pass


def isactor(type):
    try:
        if issubclass(type, Actor):
            return True
        else:
            return False
    except Exception as e:
        return False


def guard_is_structure(cls):
    if not inspect.ismethod(cls.get_identifier):
        raise NonConvertableType(
            f"Could not convert {cls.__name__} to Port: Please implement a classmethod get_identifier or decorate {cls.__name__} with register_structure(identifier='UNIQUEIDENTIFIER)"
        )


def convert_arg_to_argport(cls, **kwargs) -> ArgPort:
    # Typing Support
    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_arg_to_argport(cls.__args__[0])
            return ListArgPort.from_params(
                **kwargs,
                child=child.dict(),
            )  # We have to call the dict format here as we want the public alias __typename in the intializer

        if cls._name == "Dict":
            child = convert_arg_to_argport(cls.__args__[1])
            return DictArgPort.from_params(
                **kwargs,
                child=child.dict(),
            )  # We have to

    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool):
            return BoolArgPort.from_params(**kwargs)  # catch bool is subclass of int
        if issubclass(cls, Enum):
            return EnumArgPort.from_params(
                options={key: value._value_ for key, value in cls.__members__.items()},
                **kwargs,
            )
        if issubclass(cls, int):
            return IntArgPort.from_params(**kwargs)
        if issubclass(cls, str):
            return StringArgPort.from_params(**kwargs)

        # Custom Cases
        # Structures are a way to implement packing and unpacking of non generic types
        # These can be imported with packages and are registered automatically registered by subclassing Structure
        if hasattr(cls, "get_identifier") and inspect.ismethod(cls.get_identifier):
            print("Run here")
            return StructureArgPort.from_structure(cls, **kwargs)
        else:
            transpiler = get_transpiler_registry().get_transpiler_for_type(cls)
            return StructureArgPort.from_params(
                identifier=transpiler.structure.get_identifier(),
                transpile=transpiler.type_name,
                **kwargs,
            )

    raise ConversionError(f"No Factory for Arg Type {cls.__name__} Conversion found")


def convert_kwarg_to_kwargport(cls, **kwargs):
    # Generic Cases
    # Typing Support
    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_kwarg_to_kwargport(cls.__args__[0])
            return ListKwargPort.from_params(
                **kwargs, child=child.dict()
            )  # We have to call the dict format here as we want the public alias __typename in the intializer

        if cls._name == "Dict":
            child = convert_kwarg_to_kwargport(cls.__args__[1])
            return DictKwargPort.from_params(
                **kwargs, child=child.dict()
            )  # We have to call the dict format here as we want the public alias __typename in the intializer

    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool):
            return BoolKwargPort.from_params(
                **kwargs
            )  # bool subclass of int therefore before
        if issubclass(cls, Enum):
            return EnumKwargPort.from_params(
                options={key: value._value_ for key, value in cls.__members__.items()},
                **kwargs,
            )
        if issubclass(cls, int):
            return IntKwargPort.from_params(**kwargs)
        if issubclass(cls, str):
            return StringKwargPort.from_params(**kwargs)

        # Custom Cases
        # Structures are a way to implement packing and unpacking of non generic types
        # These can be imported with packages and are registered automatically registered by subclassing Structure
        if hasattr(cls, "get_identifier") and inspect.ismethod(cls.get_identifier):
            return StructureKwargPort.from_structure(cls, **kwargs)

        else:
            transpiler = get_transpiler_registry().get_transpiler_for_type(cls)
            return StructureKwargPort.from_params(
                identifier=transpiler.structure.get_identifier(),
                transpile=transpiler.type_name,
                **kwargs,
            )

    raise ConversionError(f"No Factory for Arg Type {cls.__name__} Conversion found")


def convert_return_to_returnport(cls, **kwargs):
    # Typing Support
    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_return_to_returnport(cls.__args__[0])
            return ListReturnPort.from_params(
                **kwargs, child=child.dict()
            )  # We have to call the dict format here as we want the public alias __typename in the intializer

        if cls._name == "Dict":
            child = convert_return_to_returnport(cls.__args__[1])
            return DictReturnPort.from_params(**kwargs, child=child.dict())  # We ha

    # Generic Cases
    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool):
            return BoolReturnPort.from_params(**kwargs)
        if issubclass(cls, int):
            return IntReturnPort.from_params(**kwargs)
        if issubclass(cls, str):
            return StringReturnPort.from_params(**kwargs)

        # Custom Cases
        # Structures are a way to implement packing and unpacking of non generic types
        # These can be imported with packages and are registered automatically registered by subclassing Structure
        if hasattr(cls, "get_identifier") and inspect.ismethod(cls.get_identifier):
            return StructureReturnPort.from_params(
                identifier=cls.get_identifier(), **kwargs
            )
        else:
            transpiler = get_transpiler_registry().get_transpiler_for_type(cls)
            return StructureReturnPort.from_params(
                identifier=transpiler.structure.get_identifier(),
                transpile=transpiler.type_name,
                **kwargs,
            )

    raise ConversionError(f"No Factory for Arg Type {cls.__name__} Conversion found")


def define(function, widgets={}, allow_empty_doc=False) -> Node:
    """Define

    Define a functions in the context of arnheim and
    return it as a Node. Attention this node is not yet
    hosted on Arkitekt (doesn't have an id). So make sure
    to save this node before calling it anywhere

    Args:
        function (): The function you want to define
    """

    is_generator = inspect.isasyncgenfunction(function) or inspect.isgeneratorfunction(
        function
    )

    sig = signature(function)
    # Generate Args and Kwargs from the Annotation
    args = []
    kwargs = []
    returns = []

    function_inputs = sig.parameters
    for key, value in function_inputs.items():
        widget = widgets.get(key, None)
        cls = value.annotation

        if value.default == Parameter.empty:
            # This Parameter is an Argument
            args.append(convert_arg_to_argport(cls, widget=widget, key=key))
        else:
            kwargs.append(
                convert_kwarg_to_kwargport(
                    cls, widget=widget, key=key, default=value.default
                )
            )

    function_output = sig.return_annotation
    try:
        # Raises type error if we use it with a class but needed here because typing is actually not a class but an Generic Alias :rolling_eyes::
        if function_output._name == "Tuple":
            for cls in function_output.__args__:
                returns.append(convert_return_to_returnport(cls))

        returns.append(
            convert_return_to_returnport(cls)
        )  # Other types will be converted to normal lists and shit

    except AttributeError:
        if function_output.__name__ != "_empty":  # Is it not empty
            returns.append(convert_return_to_returnport(function_output))

    # Documentation Parsing

    # Docstring Parser to help with descriptions
    docstring = parse(function.__doc__)
    if docstring.long_description is None:
        assert (
            allow_empty_doc is not False
        ), f"We don't allow empty documentation for function {function.__name__}. Please Provide"
        logger.warn(
            f"Allowing empty Documentatoin. Please consider providing a documentation for function {function.__name__}"
        )

    name = docstring.short_description or function.__name__
    interface = inflection.underscore(function.__name__)  # convert this to camelcase
    description = docstring.long_description or "No Description"

    doc_param_map = {
        param.arg_name: {
            "description": param.description
            if not param.description.startswith("[")
            else None,
        }
        for param in docstring.params
    }

    if docstring.returns:
        return_description = docstring.returns.description
        seperated_list = return_description.split(",")
        assert len(returns) == len(
            seperated_list
        ), f"Length of Description and Returns not Equal: If you provide a Return Annotation make sure you seperate the description for each port with ',' Return Description {return_description} Returns: {returns}"
        for index, doc in enumerate(seperated_list):
            returns[index].description = doc

    # TODO: Update with documentatoin.... (Set description for portexample)
    for port in args:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description

    for port in kwargs:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description

    return Node(
        **{
            "name": name,
            "interface": interface,
            "package": "test",  # TODO: IMplement correctly
            "description": description,
            "args": [arg.dict() for arg in args],  # exclude={"typename"} for input?
            "kwargs": [kwarg.dict() for kwarg in kwargs],
            "returns": [re.dict() for re in returns],
            "type": NodeType.GENERATOR if is_generator else NodeType.FUNCTION,
        }
    )


async def async_none(message):
    return None


def actify(
    function_or_actor,
    bypass_shrink=False,
    bypass_expand=False,
    transpilers: Dict[str, Transpiler] = None,
    on_provide=None,
    on_unprovide=None,
    actor_name=None,
    **params,
) -> Callable[[], Actor]:
    if isactor(function_or_actor):
        return function_or_actor

    actor_name = (
        actor_name or f"GeneratedActor{function_or_actor.__name__.capitalize()}"
    )

    is_coroutine = inspect.iscoroutinefunction(function_or_actor)
    is_asyncgen = inspect.isasyncgenfunction(function_or_actor)

    is_generatorfunction = inspect.isgeneratorfunction(function_or_actor)
    is_function = inspect.isfunction(function_or_actor)

    actor_attributes = {
        "assign": function_or_actor,
        "expand_inputs": not bypass_expand,
        "shrink_outputs": not bypass_shrink,
        "transpilers": transpilers,
        "on_provide": on_provide if on_provide else async_none,
        "on_unprovide": on_unprovide if on_unprovide else async_none,
    }

    if is_coroutine:
        return lambda: FunctionalFuncActor(**actor_attributes)
    elif is_asyncgen:
        return lambda: FunctionalGenActor(**actor_attributes)
    elif is_generatorfunction:
        return lambda: FunctionalThreadedGenActor(**actor_attributes)
    elif is_function:
        return lambda: FunctionalThreadedFuncActor(**actor_attributes)
    else:
        raise NotImplementedError("No way of converting this to a function")



from arkitekt.actors.functional import FunctionalFuncActor
import logging
from koil.koil import Koil
from koil.qt import TimeoutFutureWrapper
from qtpy.QtCore import QObject


logger = logging.getLogger(__name__)


class ActorSignals(QObject):
    def __init__(self, timeout=300) -> None:
        self.assign = TimeoutFutureWrapper(timeout=timeout)
        self.on_provide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)
        self.on_unprovide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)


class QtActor(FunctionalFuncActor, QObject):
    def __init__(
        self,
        *args,
        qt_assign=None,
        qt_on_provide=None,
        qt_on_unprovide=None,
        timeout=300,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)

        self.qt_assign = qt_assign
        self.qt_on_provide = qt_on_provide
        self.qt_on_unprovide = qt_on_unprovide

        self.signals = ActorSignals()

        if self.qt_assign is not None:
            self.signals.assign.call.connect(
                self.assign_function_in_qtloop
            )  # insignal will be set dymaically by qt agent

        if self.qt_on_provide is not None:
            self.signals.on_provide.connect(self.provide_function_in_qtloop)

        if self.qt_on_unprovide is not None:
            self.signals.on_unprovide.connect(self.unprovide_function_in_qtloop)

    def provide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = (
                self.qt_on_provide(*args, **kwargs) if self.qt_on_provide else None
            )
            self.signals.on_provide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_provide.reject.emit(reference, e)

    def unprovide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = (
                self.qt_on_unprovide(*args, **kwargs) if self.qt_on_unprovide else None
            )
            self.signals.on_unprovide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_unprovide.reject.emit(reference, e)

    def assign_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_assign(*args, **kwargs)
            self.signals.assign.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.assign.reject.emit(reference, e)

    async def assign(self, *args, **kwargs):
        return await self.signals.assign.acall(*args, **kwargs)

    async def on_provide(self, *args, **kwargs):
        return await self.signals.on_provide.acall(*args, **kwargs)

    async def on_unprovide(self, *args, **kwargs):
        return await self.signals.on_unprovide.acall(*args, **kwargs)



from .unassign import UnassignMessage
from .unassign_critical import UnassignCriticalMessage
from .unassign_log import UnassignLogMessage
from .bounced_unassign import BouncedUnassignMessage
from .unassign_done import UnassignDoneMessage
from .bounced_forwarded_unassign import BouncedForwardedUnassignMessage


from pydantic.main import BaseModel
from ....messages.types import PROVIDE, PROVIDE_DONE
from ....messages.base import (
    MessageDataModel,
    MessageMetaExtensionsModel,
    MessageMetaModel,
    MessageModel,
)
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]


class MetaModel(MessageMetaModel):
    type: str = PROVIDE_DONE
    extensions: Optional[MetaExtensionsModel]


class DataModel(MessageDataModel):
    debug: bool = False  # Sets the


class ProvideDoneMessage(MessageModel):
    data: DataModel
    meta: MetaModel



from pydantic.main import BaseModel
from ....messages.types import ASSIGN, PROVIDE, UNASSIGN
from ....messages.base import (
    MessageDataModel,
    MessageMetaExtensionsModel,
    MessageMetaModel,
    MessageModel,
)
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    with_progress: bool = False


class MetaModel(MessageMetaModel):
    """The reference of the metamodel representats the assignation on the platform"""

    type: str = UNASSIGN
    extensions: Optional[MetaExtensionsModel]


class DataModel(MessageDataModel):
    assignation: str
    provision: str


class UnassignMessage(MessageModel):
    data: DataModel
    meta: MetaModel



from asyncio.tasks import create_task
from arkitekt.agents.standard import StandardAgent
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.messages.postman.provide.provide_transition import (
    ProvideMode,
    ProvideState,
    ProvideTransitionMessage,
)
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from arkitekt.schema.params import TemplateParams
from fakts import config
from herre.wards.base import WardException
from arkitekt.actors.actify import actify, define
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import (
    BouncedUnprovideMessage,
)
from arkitekt.messages.base import MessageDataModel, MessageModel
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.agents.base import Agent, AgentException
import logging
from herre.console import get_current_console

logger = logging.getLogger(__name__)


class AppAgent(StandardAgent):
    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Running Actors indexed by their ID
        self.runningActors: Dict[str, Actor] = {}
        self.runningTasks: Dict[str, asyncio.Task] = {}

        self.templatedUnqueriedNodes: List[
            Tuple[dict, Callable]
        ] = []  # dict are queryparams for the node
        self.templatedNodes: List[
            Tuple[Node, Callable]
        ] = []  # node is already saved and has id
        self.templatedNewNodes: List[
            Tuple[Node, Callable]
        ] = []  # Node is not saved and has undefined id

        self.approvedTemplates: List[
            Tuple[Template, Callable]
        ] = []  # Template is approved

        # IMportant Maps
        self.templateActorBuilderMap = {}
        self.templateTemplatesMap = {}

    async def on_transport_about_to_connect(self):
        await self.approve_nodes_and_templates()
        return

    async def on_transport_connected(self):
        logger.info(f"Hosting {self.templateActorBuilderMap.keys()}")

    def on_task_done(self, future):
        logger.debug(f"Actor ended with state {future}")

    async def on_bounced_provide(self, message: BouncedProvideMessage):

        if message.data.template in self.templateActorBuilderMap:
            if message.meta.reference not in self.runningActors:
                # Didn not exist before
                actor = self.templateActorBuilderMap[
                    message.data.template
                ]()  # creating out little Actor
                self.runningActors[message.meta.reference] = actor
                task = create_task(actor.arun(message, self))
                task.add_done_callback(self.on_task_done)
                self.runningTasks[message.meta.reference] = task

            else:
                if self.strict:
                    raise AgentException(
                        "Already Running Provision Received Again. Right now causing Error. Might be omitted"
                    )
                again_provided = ProvideTransitionMessage(
                    data={
                        "message": "Provision was running on this Instance. Probably a freaking race condition",
                        "state": ProvideState.ACTIVE,
                        "mode": ProvideMode.DEBUG
                        if self.config.debug
                        else ProvideMode.PRODUCTION,
                    },
                    meta={
                        "extensions": message.meta.extensions,
                        "reference": message.meta.reference,
                    },
                )
                await self.transport.forward(again_provided)

        else:
            raise AgentException("No approved actors for this template")

    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        if message.data.provision not in self.runningActors:
            raise AgentException(
                "Already Running Provision Received Again. Right now causing Error. Might be omitted"
            )
        actor = self.runningActors[message.data.provision]

        logger.info(f"Cancelling {actor}")
        self.runningTasks[message.data.provision].cancel()

    async def on_bounced_assign(self, message: BouncedForwardedAssignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)
        else:
            if self.strict:
                raise AgentException("Received Assignment for not running Provision")

    async def on_bounced_unassign(self, message: BouncedForwardedUnassignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)

        else:
            if self.strict:
                raise AgentException("Received Assignment for not running Provision")
            logger.info("We didnt have this assignment, setting Cancellation anyways")
            await self.transport.forward(
                AssignCancelledMessage(
                    data={"canceller": "Fake Cancellation trough Provider"},
                    meta={"reference": message.data.assignation},
                )
            )

    async def approve_nodes_and_templates(self):

        if self.templatedUnqueriedNodes:
            for query_params, defined_actor, params in self.templatedUnqueriedNodes:
                try:
                    arkitekt_node = await Node.asyncs.get(**query_params)
                    self.templatedNodes.append((arkitekt_node, defined_actor, params))
                except WardException as e:
                    logger.exception(e)
                    if self.strict:
                        raise AgentException(
                            f"Couldn't find Node for query {query_params}"
                        ) from e

        if self.templatedNewNodes:
            for defined_node, defined_actor, params in self.templatedNewNodes:
                # Defined Node are nodes that are not yet reflected on arkitekt (i.e they dont have an instance
                # id so we are trying to send them to arkitekt)
                try:
                    arkitekt_node = await Node.asyncs.create(
                        **defined_node.dict(as_input=True)
                    )
                    self.templatedNodes.append((arkitekt_node, defined_actor, params))
                except WardException as e:
                    logger.exception(e)
                    if self.strict:
                        raise AgentException(
                            f"Couldn't create Node for defintion {defined_node}"
                        ) from e

        if self.templatedNodes:
            # This is an arkitekt Node and we can generate potential Templates
            for arkitekt_node, defined_actor, params in self.templatedNodes:
                try:  # Parse the parameters for template creation
                    version = params.get("version", "main")

                    arkitekt_template = await Template.asyncs.create(
                        node=arkitekt_node,
                        params=TemplateParams(**params),
                        version=version,
                    )
                    self.approvedTemplates.append(
                        (arkitekt_template, defined_actor, params)
                    )
                except WardException as e:
                    logger.exception(e)
                    if self.strict:
                        raise AgentException(
                            f"Couldn't approve template for node {arkitekt_node}"
                        ) from e

        if self.approvedTemplates:

            for arkitekt_template, defined_actor, params in self.approvedTemplates:

                # Generating Maps for Easy access
                self.templateActorBuilderMap[arkitekt_template.id] = defined_actor
                self.templateTemplatesMap[arkitekt_template.id] = arkitekt_template

                if self.panel:
                    self.panel.add_to_actor_map(arkitekt_template, defined_actor)
